/* [scarpaz] 
 * A linux system programming example, 
 * illustrating the use of shared memory and signals.
 */

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/shm.h>
#include <sys/stat.h>
#include <unistd.h>
#include <signal.h>
#include <math.h>


typedef struct 
{
  float angle, sin, cos;
  int angle_ready, sin_ready, cos_ready;
} calculation;


int           segment_id    = 0;
calculation * shared_memory = NULL;
char        * whoami        = "controller";


/* type definition for the process worker function */
typedef int    work_func(void * parameterp);


void fail(char * message)
{
  fprintf(stderr, message);
  abort();
}


void signal_handler(int i)
{
  if (strcmp(whoami,"sin")==0)
    shared_memory->sin = sin(shared_memory->angle);
  else if (strcmp(whoami,"cos")==0)
    shared_memory->cos = cos(shared_memory->angle);
  else
    fail("Illegal whoami variable.");
}


int proc(void * parameterp)
{
  int i;
  /* Install signal handler. */
  signal(SIGUSR1, signal_handler);  
  /* Wait for a signal. */
  pause();
  printf("%s> Work completed. Returning.\n", whoami);
  return 1;
}


/*

*/
int start(work_func * functionp, void * parameterp)
{
  pid_t child_pid;
  
  if (child_pid = fork())
      return child_pid;
  else 
    {
      /* We are in the new process. We set our local math_funcp according to received parameter.*/
      printf("%s> Started.\n", parameterp);      
      whoami = parameterp;	
      exit(functionp(parameterp));
    }
}



int main()
{
  int cos_pid, sin_pid, pid, status, i;    

  /* creating a new private shared memory segment */
  segment_id = shmget(IPC_PRIVATE, sizeof(calculation), IPC_CREAT | IPC_EXCL | S_IRUSR | S_IWUSR);  
  /* -to learn more on IPC_* masks, do 'man 2 shmget',
     -to learn more on S_* masks, do 'man 2 stat'
  */
  printf("%s> Allocated segment ID = %i.\n", whoami, segment_id);
  
  /* attaching the shared memory segment to a location of our current addressing space;
   * shared_memory now points to a structure stored in the shared segment;
   */
  shared_memory = (calculation*) shmat(segment_id, 0, 0);
  printf("%s> Segment attached at address %p.\n", whoami, shared_memory);

  /* starting the cosine calculation process */
  cos_pid = start(&proc, "cos");
  /* starting the sine calculation process */
  sin_pid = start(&proc, "sin");

  /* clearing the calculation structure*/
  memset(shared_memory, 0, sizeof(calculation));
  /* setting the angle for which sine and cosine is to be calculated */
  shared_memory->angle  = 0.35967;

  /* sending to each process a user signal */
  kill(cos_pid, SIGUSR1);
  kill(sin_pid, SIGUSR1);  
  
  for (i=0; i<2; i++)
    {
      /* waiting for each process to terminate */
      pid = wait(&status);  
      printf("%s> Child process [PID %i] terminated, " 
	     "normal termination = %i, exit code = %i.\n", 
	     whoami, pid, WIFEXITED(status), WEXITSTATUS(status));
    }  

  printf("%s> All children complete.\n"
	 "\tangle       = %g\n"
         "\tsin(angle)  = %g\n"
         "\tcos(angle)  = %g\n"
         "\tsin^2+cos^2 = %g\n", 
	 whoami, shared_memory->angle, shared_memory->sin, shared_memory->cos,
	 shared_memory->sin*shared_memory->sin + shared_memory->cos*shared_memory->cos
	 );

  /* detaching the shared memory segment from the addressing space */
  shmdt(shared_memory);
  /* marking the shared memory segment for destruction */
  shmctl(segment_id, IPC_RMID, 0);
  
  return 0;
}
